#include "headers.h"

static void do_reorder(queue_t *q, send_t **qsends, int no_qsends, int rm) {
	int i,j,found,to_alloc;

	PTRARR_DEL(&qsends,&no_qsends,NULL);
	
	for(i = 0; i < q->no_qsends; ++i) {
		for(j = 0, found = 0; j < no_qsends && !found; ++j) {
			if(q->qsends[i] == qsends[j])
				found = 1;
		}
		if(!found && rm) {
			free_send(q->qsends[i]);
			q->qsends[i] = NULL;
		} else if(!found) {
			PTRARR_ADD(&qsends,&no_qsends,q->qsends[i]);
		}
	}

	to_alloc = no_qsends;
	for(i = 0; i < no_qsends; ++i) {
		for(j = 0, found = 0; j < q->no_qsends && !found; ++j) {
			if(qsends[i] == q->qsends[j])
				found = 1;
		}
		if(!found) {
			qsends[i] = NULL;
			to_alloc--;
		}
	}

	xfree(q->qsends);
	PTRARR_ALLOC(&(q->qsends),&(q->no_qsends),to_alloc);

	for(i = 0, j = 0; i < no_qsends; ++i) {
		if(qsends[i] != NULL)
			q->qsends[j++] = qsends[i];
	}
}

static void reorder_queues(SV **sp, SV **mark, I32 ax, I32 items,
		queue_t *q, int rm) {
	send_t **qsends;
	int i, no_qsends;;
	
	no_qsends = items - 1;
	qsends = xalloc(sizeof(send_t *)*(no_qsends+1));
	
	for(i = 0; i < no_qsends; ++i) {
		HV *hv = hvref(ST(i+1));
		qsends[i] = NULL;
		if(hv != NULL) {
			SV **sv = hv_fetch(hv,"_ptr",4,0);
			if(sv != NULL)
				qsends[i] = (send_t *)SvIV(*sv);
		}
	}
	
	do_reorder(q,qsends,no_qsends,rm);
	xfree(qsends);
}

MODULE = Arisa::Queue  PACKAGE = Arisa
PROTOTYPES: ENABLE

Arisa::Queue
queue_find(name = NULL)
	char *name
CODE:
	RETVAL = queue_find(name);
OUTPUT:
	RETVAL

void
queues()
PREINIT:
	int i;
PPCODE:
	LOCK(global);
	for(i = 0; i < global->no_queues; ++i) {
		queue_t *q = global->queues[i];
		LOCK(q);
		if(q->deleted == 0)
			XPUSHs(sv_2mortal(
				create_ptr_object("Arisa::Queue",q)));
		UNLOCK(q);
	}
	UNLOCK(global);

Arisa::Queue
queue_add(name = NULL, pool)
	char *name
	Arisa::Pool pool
PREINIT:
	interface_t *intf;
	pool_t *p;
CODE:
	RETVAL = NULL;
	PTR2SPLIT(pool,pool_t,p,interface_t,intf);
	if(name != NULL && pool_valid(intf,p)) {
		queue_t *q;
		
		q = alloc_queue();
		q->name = xstrdup(name);
		q->pool = p;
		
		if(queue_add(q) == 0) {
			LOCK(q);
			LOCK(q->pool);
			if(q->pool->deleted != 0) {
				UNLOCK(q->pool);
				UNLOCK(q);
				queue_del(q);
			} else {
				UNLOCK(q->pool);
				UNLOCK(q);
				RETVAL = q;
			}
		} else {
			free_queue(q);
		}
	}
OUTPUT:
	RETVAL

	
#*******************************
MODULE = Arisa::Queue  PACKAGE = Arisa::Queue
#*******************************

SV*
check_users(queue)
	Arisa::Queue queue
CODE:
	RETVAL = &PL_sv_no;
	if(queue_valid(queue)) {
		if(queue_check_users(queue) == 0)
			RETVAL = &PL_sv_yes;
	}
OUTPUT:
	RETVAL

SV*
delete(queue)
	Arisa::Queue queue
CODE:
	RETVAL = &PL_sv_no;
	if(queue_valid(queue)) {
		if(queue_del(queue) == 0)
			RETVAL = &PL_sv_yes;
	}
OUTPUT:
	RETVAL

SV*
name(queue, new_name = NULL)
	Arisa::Queue queue
	char *new_name
CODE:
	RETVAL = &PL_sv_no;
	if(queue_valid(queue)) {
		if(new_name != NULL) {
			if(queue_rename(queue,new_name) == 0)
				RETVAL = &PL_sv_yes;
		} else {
			LOCK(queue);
			RETVAL = new_pv(queue->name);
			UNLOCK(queue);
		}
	}
OUTPUT:
	RETVAL

void
info(q)
	Arisa::Queue q
PREINIT:
	HV *hv = NULL;
PPCODE:
	if(queue_valid(q)) {
		hv = newHV();
		LOCK(q);
		hv_store(hv,"last_notify",11,newSVuv(q->last_notify),0);
		hv_store(hv,"last_dequeue",12,newSVuv(q->last_dequeue),0);
		hv_store(hv,"last_checked",12,newSVuv(q->last_checked),0);
		UNLOCK(q);
	}
	RET_REF(hv);

SV*
move(q, from_pos, to_pos)
	Arisa::Queue q
	int from_pos
	int to_pos
CODE:
	RETVAL = &PL_sv_no;
	if(queue_valid(q)) {
		if(queue_send_move(q,from_pos,to_pos) == 0)
			RETVAL = &PL_sv_yes;
	}
OUTPUT:
	RETVAL

int
notify_users(q)
	Arisa::Queue q
CODE:
	RETVAL = -1;
	if(queue_valid(q))
		RETVAL = queue_notify_users(q);
OUTPUT:
	RETVAL
	
Arisa::Pool
pool(q, ...)
	Arisa::Queue q
CODE:
	RETVAL.major = NULL;
	RETVAL.minor = NULL;
	if(queue_valid(q)) {
		int i,j;
		
		if(items > 1) {
			ptr2_t ptr;
			fetch_object_ptr2(ST(1),&ptr);
			if(pool_valid(ptr.minor,ptr.major)) {
				pool_t *tmp;
				LOCK(q);
				tmp = q->pool;
				q->pool = ptr.major;
				LOCK(q->pool);
				if(q->pool->deleted != 0) {
					UNLOCK(q->pool);
					q->pool = tmp;
				} else {
					UNLOCK(q->pool);
				}
				UNLOCK(q);
			}
		}
		
		LOCK(q);
		RETVAL.major = q->pool;
		UNLOCK(q);
		
		LOCK(global);
		for(i = 0; i < global->no_interfaces && 
				RETVAL.minor == NULL; ++i) {
			interface_t *intf = global->interfaces[i];
			LOCK(intf);
			if(intf->type == INTERFACE_SEND) {
				for(j = 0; j < intf->send.no_pools &&
						RETVAL.minor == NULL; ++j) {
					if(intf->send.pools[j] == RETVAL.major)
						RETVAL.minor = intf;
				}
			}
			UNLOCK(intf);
		}
		UNLOCK(global);
	}
OUTPUT:
	RETVAL

void
queues(q)
	Arisa::Queue q
PPCODE:
	if(queue_valid(q)) {
		int i;
		LOCK(q);
		for(i = 0; i < q->no_qsends; ++i) {
			LOCK(q->qsends[i]);
			XPUSHs(sv_2mortal(newRV_noinc(
				(SV*)send_to_hv(q->qsends[i]))));
			UNLOCK(q->qsends[i]);
		}
		UNLOCK(q);
	}
	
int
remove(q, ...)
	Arisa::Queue q
CODE:
	RETVAL = 0;
	if(queue_valid(q)) {
		LOCK(q);
		if(items == 2) {
			if(SvTYPE(ST(1)) == SVt_IV) {
				int no = SvIV(ST(1));
				if(queue_remove_no(q,no) == 0)
					RETVAL = 1;
			}
		}
		if(RETVAL == 0) {
			network_t *net	= SvPTRarg("network");
			char *host	= SvPVarg("host");
			char *nick	= SvPVarg("nick");
			char *file	= SvPVarg("file");
			char *label	= SvPVarg("label");
			void *ptr	= (void *)SvIVarg("_ptr");
			
			RETVAL = queue_remove(q,host,net,nick,file,label,ptr);
		}
		UNLOCK(q);
	}
OUTPUT:
	RETVAL

void
reorder(q, ...)
	Arisa::Queue q
PPCODE:
	if(queue_valid(q)) {
		int i;
		
		LOCK(q);

		reorder_queues(sp,mark,ax,items,q,0);
		
		for(i = 0; i < q->no_qsends; ++i) {
			LOCK(q->qsends[i]);
			XPUSHs(sv_2mortal(newRV_noinc(
				(SV*)send_to_hv(q->qsends[i]))));
			UNLOCK(q->qsends[i]);
		}
		
		UNLOCK(q);
	}

void
reorder_and_remove(q, ...)
	Arisa::Queue q
PPCODE:
	if(queue_valid(q)) {
		int i;
		
		LOCK(q);
		
		reorder_queues(sp,mark,ax,items,q,1);
		
		for(i = 0; i < q->no_qsends; ++i) {
			LOCK(q->qsends[i]);
			XPUSHs(sv_2mortal(newRV_noinc(
				(SV*)send_to_hv(q->qsends[i]))));
			UNLOCK(q->qsends[i]);
		}
		
		UNLOCK(q);
	}

int
send(q, ...)
	Arisa::Queue q
CODE:
	RETVAL = 0;
	if(queue_valid(q)) {
		if(items == 2) {
			if(SvTYPE(ST(1)) == SVt_IV) {
				int no = SvIV(ST(1));
				if(queue_send_force_no(q,no) == 0)
					RETVAL = 1;
			}
		}
		if(RETVAL == 0) {
			network_t *net	= SvPTRarg("network");
			char *host	= SvPVarg("host");
			char *nick	= SvPVarg("nick");
			char *file	= SvPVarg("file");
			char *label	= SvPVarg("label");
			void *ptr	= (void *)SvIVarg("_ptr");

			RETVAL = queue_send_force(q,host,net,nick,file,label,ptr);
		}
	}
OUTPUT:
	RETVAL
	
void
settings(queue, ...)
	Arisa::Queue queue
PREINIT:
	HV *hv = NULL;
PPCODE:
	if(queue_valid(queue)) {
		hv = newHV();
		perl_apply_settings(queue_apply_setting,queue);
		perl_read_settings(queue_read_settings,queue,hv);
	}
	RET_REF(hv);

SV*
swap(q, pos1, pos2)
	Arisa::Queue q
	int pos1
	int pos2
CODE:
	RETVAL = &PL_sv_no;
	if(queue_valid(q)) {
		if(queue_send_swap(q,pos1,pos2) == 0)
			RETVAL = &PL_sv_yes;
	}
OUTPUT:
	RETVAL
